==> token-validation-ktor-1.3.4-sources.jar: no/nav/security/token/support/ktor/TokenSupportAuthenticationProvider.kt <==
package no.nav.security.token.support.ktor

import io.ktor.application.*
import io.ktor.auth.*
import io.ktor.config.*
import io.ktor.http.*
import io.ktor.request.*
import io.ktor.response.*
import io.ktor.util.*
import no.nav.security.token.support.core.configuration.IssuerProperties
import no.nav.security.token.support.core.configuration.MultiIssuerConfiguration
import no.nav.security.token.support.core.configuration.ProxyAwareResourceRetriever
import no.nav.security.token.support.core.context.TokenValidationContext
import no.nav.security.token.support.core.context.TokenValidationContextHolder
import no.nav.security.token.support.core.http.HttpRequest
import no.nav.security.token.support.core.validation.JwtTokenAnnotationHandler
import no.nav.security.token.support.core.validation.JwtTokenValidationHandler
import org.slf4j.LoggerFactory
import java.net.URL

data class TokenValidationContextPrincipal(val context: TokenValidationContext) : Principal

@KtorExperimentalAPI
private val log = LoggerFactory.getLogger(TokenSupportAuthenticationProvider::class.java.name)

@KtorExperimentalAPI
class TokenSupportAuthenticationProvider(
    providerConfig: ProviderConfiguration,
    applicationConfig: ApplicationConfig,
    resourceRetriever: ProxyAwareResourceRetriever
) : AuthenticationProvider(providerConfig) {

    @Deprecated("Provider should be built using configuration that need to be passed via constructor instead.")
    constructor(
        name: String?,
        config: ApplicationConfig,
        resourceRetriever: ProxyAwareResourceRetriever
    ): this(ProviderConfiguration(name),config, resourceRetriever)

    internal val jwtTokenValidationHandler: JwtTokenValidationHandler
    internal val jwtTokenExpiryThresholdHandler: JwtTokenExpiryThresholdHandler

    init {
        val issuerPropertiesMap: Map<String, IssuerProperties> = applicationConfig.asIssuerProps()
        jwtTokenValidationHandler = JwtTokenValidationHandler(
            MultiIssuerConfiguration(issuerPropertiesMap, resourceRetriever)
        )

        val expiryThreshold: Int = applicationConfig.propertyOrNull("no.nav.security.jwt.expirythreshold")?.getString()?.toInt() ?: -1
        jwtTokenExpiryThresholdHandler = JwtTokenExpiryThresholdHandler(expiryThreshold)
    }

    class ProviderConfiguration internal constructor(name: String?): AuthenticationProvider.Configuration(name)
}

@KtorExperimentalAPI
fun Authentication.Configuration.tokenValidationSupport(
    name: String? = null,
    config: ApplicationConfig,
    requiredClaims: RequiredClaims? = null,
    additionalValidation: ((TokenValidationContext) -> Boolean)? = null,
    resourceRetriever: ProxyAwareResourceRetriever = ProxyAwareResourceRetriever(
        System.getenv("HTTP_PROXY")?.let { URL(it) }
    )
) {
    val provider = TokenSupportAuthenticationProvider(
        TokenSupportAuthenticationProvider.ProviderConfiguration(name),
        config,
        resourceRetriever
    )
    provider.pipeline.intercept(AuthenticationPipeline.RequestAuthentication) { context ->
        val tokenValidationContext = provider.jwtTokenValidationHandler.getValidatedTokens(
            JwtTokenHttpRequest(call.request.cookies, call.request.headers)
        )
        try {
            if (tokenValidationContext.hasValidToken()) {
                if (requiredClaims != null) {
                    RequiredClaimsHandler(InternalTokenValidationContextHolder(tokenValidationContext)).handleRequiredClaims(
                        requiredClaims
                    )
                }
                if (additionalValidation != null) {
                    if (!additionalValidation(tokenValidationContext)) {
                        throw AdditionalValidationReturnedFalse()
                    }
                }
                provider.jwtTokenExpiryThresholdHandler.addHeaderOnTokenExpiryThreshold(call, tokenValidationContext)
                context.principal(TokenValidationContextPrincipal(tokenValidationContext))
                return@intercept
            }
        } catch (e: Throwable) {
            val message = e.message ?: e.javaClass.simpleName
            log.debug("Token verification failed: {}", message)
        }
        context.challenge("JWTAuthKey", AuthenticationFailedCause.InvalidCredentials) {
            call.respond(UnauthorizedResponse())
            it.complete()
        }
    }
    register(provider)
}


data class RequiredClaims(val issuer: String, val claimMap: Array<String>, val combineWithOr: Boolean = false)

data class IssuerConfig(
    val name: String,
    val discoveryUrl: String,
    val acceptedAudience: List<String>,
    val cookieName: String? = null
)

@KtorExperimentalAPI
class TokenSupportConfig(vararg issuers: IssuerConfig) : MapApplicationConfig(
    *(issuers.mapIndexed { index, issuerConfig ->
        listOf(
            "no.nav.security.jwt.issuers.$index.issuer_name" to issuerConfig.name,
            "no.nav.security.jwt.issuers.$index.discoveryurl" to issuerConfig.discoveryUrl,
            "no.nav.security.jwt.issuers.$index.accepted_audience" to issuerConfig.acceptedAudience.joinToString(",")//,
        ).let {
            if (issuerConfig.cookieName != null) {
                it.plus("no.nav.security.jwt.issuers.$index.cookie_name" to issuerConfig.cookieName)
            } else {
                it
            }
        }
    }.flatMap { it }.plus("no.nav.security.jwt.issuers.size" to issuers.size.toString()).toTypedArray())
)

private class InternalTokenValidationContextHolder(private var tokenValidationContext: TokenValidationContext) :
    TokenValidationContextHolder {
    override fun getTokenValidationContext() = tokenValidationContext
    override fun setTokenValidationContext(tokenValidationContext: TokenValidationContext?) {
        this.tokenValidationContext = tokenValidationContext!!
    }
}

internal class AdditionalValidationReturnedFalse : RuntimeException()

internal class RequiredClaimsException(message: String, cause: Exception) : RuntimeException(message, cause)
internal class RequiredClaimsHandler(tokenValidationContextHolder: TokenValidationContextHolder) :
    JwtTokenAnnotationHandler(tokenValidationContextHolder) {
    internal fun handleRequiredClaims(requiredClaims: RequiredClaims) {
        try {
            handleProtectedWithClaims(requiredClaims.issuer, requiredClaims.claimMap, requiredClaims.combineWithOr)
        } catch (e: RuntimeException) {
            throw RequiredClaimsException(e.message ?: "", e)
        }
    }
}

internal data class NameValueCookie(@JvmField val name: String, @JvmField val value: String) : HttpRequest.NameValue {
    override fun getName(): String = name
    override fun getValue(): String = value
}

internal data class JwtTokenHttpRequest(private val cookies: RequestCookies, private val headers: Headers) :
    HttpRequest {
    @KtorExperimentalAPI
    override fun getCookies() =
        cookies.rawCookies.map {
            NameValueCookie(
                it.key,
                decodeCookieValue(it.value, CookieEncoding.URI_ENCODING)
            )
        }.toTypedArray()

    override fun getHeader(name: String) = headers[name]
}

fun ApplicationConfig.asIssuerProps(): Map<String, IssuerProperties> = this.configList("no.nav.security.jwt.issuers")
    .associate { issuerConfig ->
        issuerConfig.property("issuer_name").getString() to IssuerProperties(
            URL(issuerConfig.property("discoveryurl").getString()),
            issuerConfig.property("accepted_audience").getString().split(","),
            issuerConfig.propertyOrNull("cookie_name")?.getString(),
            IssuerProperties.Validation(issuerConfig.propertyOrNull("validation.optional_claims")?.getString()?.split(",") ?: emptyList()),
            IssuerProperties.JwksCache(
                issuerConfig.propertyOrNull("jwks_cache.lifespan")?.getString()?.toLong(),
                issuerConfig.propertyOrNull("jwks_cache.refreshtime")?.getString()?.toLong()
            )
        )
    }

==> token-validation-ktor-1.3.6-sources.jar: no/nav/security/token/support/ktor/TokenSupportAuthenticationProvider.kt <==
package no.nav.security.token.support.ktor

import io.ktor.application.*
import io.ktor.auth.*
import io.ktor.config.*
import io.ktor.http.*
import io.ktor.request.*
import io.ktor.response.*
import io.ktor.util.*
import no.nav.security.token.support.core.utils.JwtTokenUtil.getJwtToken
import no.nav.security.token.support.core.configuration.MultiIssuerConfiguration
import no.nav.security.token.support.core.configuration.ProxyAwareResourceRetriever
import no.nav.security.token.support.core.context.TokenValidationContext
import no.nav.security.token.support.core.context.TokenValidationContextHolder
import no.nav.security.token.support.core.http.HttpRequest
import no.nav.security.token.support.core.validation.JwtTokenAnnotationHandler
import no.nav.security.token.support.core.validation.JwtTokenValidationHandler
import org.slf4j.LoggerFactory
import java.net.URL
import no.nav.security.token.support.core.exceptions.JwtTokenInvalidClaimException
import no.nav.security.token.support.core.exceptions.JwtTokenMissingException
import no.nav.security.token.support.core.configuration.IssuerProperties

data class TokenValidationContextPrincipal(val context: TokenValidationContext) : Principal

@KtorExperimentalAPI
private val log = LoggerFactory.getLogger(TokenSupportAuthenticationProvider::class.java.name)

@KtorExperimentalAPI
class TokenSupportAuthenticationProvider(
    providerConfig: ProviderConfiguration,
    applicationConfig: ApplicationConfig,
    resourceRetriever: ProxyAwareResourceRetriever
) : AuthenticationProvider(providerConfig) {

    @Deprecated("Provider should be built using configuration that need to be passed via constructor instead.")
    constructor(
        name: String?,
        config: ApplicationConfig,
        resourceRetriever: ProxyAwareResourceRetriever
    ): this(ProviderConfiguration(name),config, resourceRetriever)

    internal val jwtTokenValidationHandler: JwtTokenValidationHandler
    internal val jwtTokenExpiryThresholdHandler: JwtTokenExpiryThresholdHandler

    init {
        val issuerPropertiesMap: Map<String, IssuerProperties> = applicationConfig.asIssuerProps()
        jwtTokenValidationHandler = JwtTokenValidationHandler(
            MultiIssuerConfiguration(issuerPropertiesMap, resourceRetriever)
        )

        val expiryThreshold: Int = applicationConfig.propertyOrNull("no.nav.security.jwt.expirythreshold")?.getString()?.toInt() ?: -1
        jwtTokenExpiryThresholdHandler = JwtTokenExpiryThresholdHandler(expiryThreshold)
    }

    class ProviderConfiguration internal constructor(name: String?): AuthenticationProvider.Configuration(name)
}

@KtorExperimentalAPI
fun Authentication.Configuration.tokenValidationSupport(
    name: String? = null,
    config: ApplicationConfig,
    requiredClaims: RequiredClaims? = null,
    additionalValidation: ((TokenValidationContext) -> Boolean)? = null,
    resourceRetriever: ProxyAwareResourceRetriever = ProxyAwareResourceRetriever(
        System.getenv("HTTP_PROXY")?.let { URL(it) }
    )
) {
    val provider = TokenSupportAuthenticationProvider(
        TokenSupportAuthenticationProvider.ProviderConfiguration(name),
        config,
        resourceRetriever
    )
    provider.pipeline.intercept(AuthenticationPipeline.RequestAuthentication) { context ->
        val tokenValidationContext = provider.jwtTokenValidationHandler.getValidatedTokens(
            JwtTokenHttpRequest(call.request.cookies, call.request.headers)
        )
        try {
            if (tokenValidationContext.hasValidToken()) {
                if (requiredClaims != null) {
                    RequiredClaimsHandler(InternalTokenValidationContextHolder(tokenValidationContext)).handleRequiredClaims(
                        requiredClaims
                    )
                }
                if (additionalValidation != null) {
                    if (!additionalValidation(tokenValidationContext)) {
                        throw AdditionalValidationReturnedFalse()
                    }
                }
                provider.jwtTokenExpiryThresholdHandler.addHeaderOnTokenExpiryThreshold(call, tokenValidationContext)
                context.principal(TokenValidationContextPrincipal(tokenValidationContext))
                return@intercept
            }
        } catch (e: Throwable) {
            val message = e.message ?: e.javaClass.simpleName
            log.debug("Token verification failed: {}", message)
        }
        context.challenge("JWTAuthKey", AuthenticationFailedCause.InvalidCredentials) {
            call.respond(UnauthorizedResponse())
            it.complete()
        }
    }
    register(provider)
}


data class RequiredClaims(val issuer: String, val claimMap: Array<String>, val combineWithOr: Boolean = false)

data class IssuerConfig(
    val name: String,
    val discoveryUrl: String,
    val acceptedAudience: List<String>,
    val cookieName: String? = null
)

@KtorExperimentalAPI
class TokenSupportConfig(vararg issuers: IssuerConfig) : MapApplicationConfig(
    *(issuers.mapIndexed { index, issuerConfig ->
        listOf(
            "no.nav.security.jwt.issuers.$index.issuer_name" to issuerConfig.name,
            "no.nav.security.jwt.issuers.$index.discoveryurl" to issuerConfig.discoveryUrl,
            "no.nav.security.jwt.issuers.$index.accepted_audience" to issuerConfig.acceptedAudience.joinToString(",")//,
        ).let {
            if (issuerConfig.cookieName != null) {
                it.plus("no.nav.security.jwt.issuers.$index.cookie_name" to issuerConfig.cookieName)
            } else {
                it
            }
        }
    }.flatMap { it }.plus("no.nav.security.jwt.issuers.size" to issuers.size.toString()).toTypedArray())
)

private class InternalTokenValidationContextHolder(private var tokenValidationContext: TokenValidationContext) :
    TokenValidationContextHolder {
    override fun getTokenValidationContext() = tokenValidationContext
    override fun setTokenValidationContext(tokenValidationContext: TokenValidationContext?) {
        this.tokenValidationContext = tokenValidationContext!!
    }
}

internal class AdditionalValidationReturnedFalse : RuntimeException()

internal class RequiredClaimsException(message: String, cause: Exception) : RuntimeException(message, cause)
internal class RequiredClaimsHandler(val tokenValidationContextHolder: TokenValidationContextHolder) :
    JwtTokenAnnotationHandler(tokenValidationContextHolder) {
    internal fun handleRequiredClaims(requiredClaims: RequiredClaims) {
        try {
			val jwtToken = getJwtToken(requiredClaims.issuer, tokenValidationContextHolder);
            if (jwtToken.isEmpty()) {
                throw  JwtTokenMissingException("no valid token found in validation context");
            }
            if (!handleProtectedWithClaims(requiredClaims.issuer, requiredClaims.claimMap, requiredClaims.combineWithOr,jwtToken.get())) 
                throw  JwtTokenInvalidClaimException("required claims not present in token." + requiredClaims.claimMap)
            
        } catch (e: RuntimeException) {
            throw RequiredClaimsException(e.message ?: "", e)
        }
    }
}

internal data class NameValueCookie(@JvmField val name: String, @JvmField val value: String) : HttpRequest.NameValue {
    override fun getName(): String = name
    override fun getValue(): String = value
}

internal data class JwtTokenHttpRequest(private val cookies: RequestCookies, private val headers: Headers) :
    HttpRequest {
    @KtorExperimentalAPI
    override fun getCookies() =
        cookies.rawCookies.map {
            NameValueCookie(
                it.key,
                decodeCookieValue(it.value, CookieEncoding.URI_ENCODING)
            )
        }.toTypedArray()

    override fun getHeader(name: String) = headers[name]
}

fun ApplicationConfig.asIssuerProps(): Map<String, IssuerProperties> = this.configList("no.nav.security.jwt.issuers")
    .associate { issuerConfig ->
        issuerConfig.property("issuer_name").getString() to IssuerProperties(
            URL(issuerConfig.property("discoveryurl").getString()),
            issuerConfig.property("accepted_audience").getString().split(","),
            issuerConfig.propertyOrNull("cookie_name")?.getString(),
            IssuerProperties.Validation(issuerConfig.propertyOrNull("validation.optional_claims")?.getString()?.split(",") ?: emptyList()),
            IssuerProperties.JwksCache(
                issuerConfig.propertyOrNull("jwks_cache.lifespan")?.getString()?.toLong(),
                issuerConfig.propertyOrNull("jwks_cache.refreshtime")?.getString()?.toLong()
            )
        )
    }
